"""
Contains config and settings for the repo.
"""

import importlib
import json
import logging
import os
import sys
import time
import warnings
from contextlib import ContextDecorator
from pathlib import Path
from typing import Any, Optional

import boto3
from PIL import Image, ImageFile

from spacer.exceptions import ConfigError


def filter_warnings():
    """ Filters out some verified warnings. """

    # Per discussion in https://github.com/boto/boto3/issues/454,
    # the boto package is raising a lot of warnings that it shouldn't.
    warnings.filterwarnings("ignore", category=ResourceWarning,
                            message="unclosed.*<ssl.SSLSocket.*>")
    warnings.filterwarnings("ignore", category=ResourceWarning,
                            message="unclosed.*<_io.TextIOWrapper.*>")


APP_DIR = Path(__file__).resolve().parent
REPO_DIR = APP_DIR.parent


# One way to specify settings is through a secrets.json file. Example:
# {
#     "AWS_ACCESS_KEY_ID": "...",
#     "AWS_SECRET_ACCESS_KEY": "..."
# }
SECRETS_PATH = REPO_DIR / 'secrets.json'
SECRETS = None
if SECRETS_PATH.exists():
    with open(SECRETS_PATH) as fp:
        SECRETS = json.load(fp)


# Another way is through Django settings. Example:
# SPACER = {
#     'AWS_ACCESS_KEY_ID': '...',
#     'AWS_SECRET_ACCESS_KEY': '...',
# }
SETTINGS_FROM_DJANGO: Optional[dict] = None
try:
    from django.core.exceptions import ImproperlyConfigured
except ImportError:
    pass
else:
    # This by itself shouldn't get errors.
    from django.conf import settings

    # If settings module can't be found, this gets ImproperlyConfigured.
    # If the module can be found, but the SPACER setting is absent, this
    # gets AttributeError.
    try:
        SETTINGS_FROM_DJANGO = settings.SPACER
    except (ImproperlyConfigured, AttributeError):
        pass


def get_config_detection_result():
    result = ""

    if SECRETS:
        result += "Secrets file found."
    else:
        result += "Secrets file not found."

    if SETTINGS_FROM_DJANGO:
        result += " SPACER Django setting found."
    else:
        result += " SPACER Django setting not found."

    return result


def get_config_value(key: str, default: Any = 'NO_DEFAULT') -> Any:

    def is_valid_value(value_):
        # Treat an empty string the same as not specifying a setting.
        return value_ not in ['', None]

    # Try environment variables. Each should be prefixed with 'SPACER_'.
    value = os.getenv('SPACER_' + key)
    if is_valid_value(value):
        return value

    def handle_unspecified_setting():
        if default == 'NO_DEFAULT':
            raise ConfigError(
                f"{key} setting is required."
                f" (Debug info: {get_config_detection_result()})"
            )
        return default

    # Try secrets file.
    if SECRETS:
        value = SECRETS.get(key)
        if is_valid_value(value):
            return value
        return handle_unspecified_setting()

    # Try Django settings.
    if SETTINGS_FROM_DJANGO:
        try:
            value = SETTINGS_FROM_DJANGO[key]
        except KeyError:
            return handle_unspecified_setting()
        else:
            if is_valid_value(value):
                return value
            return handle_unspecified_setting()

    return handle_unspecified_setting()


def get_s3_conn():
    """
    Returns a boto s3 connection.
    - It first looks for credentials in spacer config.
    - If not found there it will default to credentials in ~/.aws/credentials
    """
    if not all([AWS_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY]):
        raise ConfigError(
            "All AWS config variables must be specified to use S3.")

    return boto3.resource('s3',
                          region_name=AWS_REGION,
                          aws_access_key_id=AWS_ACCESS_KEY_ID,
                          aws_secret_access_key=AWS_SECRET_ACCESS_KEY)


class log_entry_and_exit(ContextDecorator):
    def __init__(self, name):
        self.name = name
        self.start_time = None

    def __enter__(self):
        self.start_time = time.time()
        logging.info('Entering: %s', self.name)

    def __exit__(self, exc_type, exc, exc_tb):
        logging.info('Exiting: %s after %f seconds.', self.name,
                     time.time() - self.start_time)


AWS_ACCESS_KEY_ID = get_config_value('AWS_ACCESS_KEY_ID', default=None)
AWS_SECRET_ACCESS_KEY = get_config_value('AWS_SECRET_ACCESS_KEY', default=None)
AWS_REGION = get_config_value('AWS_REGION', default=None)

# Filesystem directory to use for caching downloaded feature-extractor data.
# This will be used whenever there is s3 or url based extractor data.
EXTRACTORS_CACHE_DIR = get_config_value('EXTRACTORS_CACHE_DIR', default=None)

TASKS = [
    'extract_features',
    'train_classifier',
    'classify_features',
    'classify_image'
]

CLASSIFIER_TYPES = [
    'LR',
    'MLP'
]

TRAINER_NAMES = [
    'minibatch'
]

# For extractors used in unit tests.
TEST_EXTRACTORS_BUCKET = get_config_value(
    'TEST_EXTRACTORS_BUCKET', default=None)
# For other fixtures used in unit tests.
#
# At least for now, the main reason these bucket names are pulled from
# config is to not expose the bucket names used by the PySpacer core devs.
# However, since these test files are not publicly linked and need to
# live in an S3 bucket with specific filenames (specified by TEST_EXTRACTORS
# and individual tests), the tests are still onerous to set up for anyone
# besides the core devs. This should be addressed sometime.
TEST_BUCKET = get_config_value('TEST_BUCKET', default=None)
# A few other fixtures live here.
LOCAL_FIXTURE_DIR = str(APP_DIR / 'tests' / 'fixtures')

STORAGE_TYPES = [
    's3',
    'filesystem',
    'memory',
    'url'
]

MAX_IMAGE_PIXELS = get_config_value('MAX_IMAGE_PIXELS', default=10000*10000)
MAX_POINTS_PER_IMAGE = get_config_value('MAX_POINTS_PER_IMAGE', default=1000)

# The train_classifier task requires as least this many images.
MIN_TRAINIMAGES = get_config_value('MIN_TRAINIMAGES', default=10)

# Check access to select which tests to run.
HAS_CAFFE = importlib.util.find_spec("caffe") is not None


# Add margin to avoid warnings when running unit-test.
Image.MAX_IMAGE_PIXELS = MAX_IMAGE_PIXELS + 20000

# Configure Pillow to be tolerant of image files that are truncated (missing
# data from the last block).
# https://stackoverflow.com/a/23575424/
ImageFile.LOAD_TRUNCATED_IMAGES = True


CONFIGURABLE_VARS = [
    # These two variables are required if you're using AWS S3 storage,
    # unless spacer is running on an AWS instance which has been set up with
    # `aws configure`.
    'AWS_ACCESS_KEY_ID',
    'AWS_SECRET_ACCESS_KEY',
    # This is required if you're using S3 storage.
    'AWS_REGION',
    # This is required if you're loading feature extractors from a remote
    # source (S3 or URL).
    'EXTRACTORS_CACHE_DIR',
    # These are required to run certain unit tests. They're also not really
    # usable by anyone besides spacer's core devs at the moment.
    'TEST_EXTRACTORS_BUCKET',
    'TEST_BUCKET',
    # These can just be configured as needed, or left as defaults.
    'MAX_IMAGE_PIXELS',
    'MAX_POINTS_PER_IMAGE',
    'MIN_TRAINIMAGES',
]


def check():
    """
    Print values of all configurable variables.
    """
    print(get_config_detection_result())

    for var_name in CONFIGURABLE_VARS:
        # Get the var_name attribute in the current module
        var_value = getattr(sys.modules[__name__], var_name)

        if var_value and '_KEY' in var_name:
            # Treat this as a sensitive value; don't print the entire thing
            value_display = f'{var_value[:6]} ... {var_value[-6:]}'
        else:
            value_display = str(var_value)

        print(f"{var_name}: {value_display}")



"""
Defines feature-extractor ABC; implementations; and factory.
"""

from __future__ import annotations
import abc
import hashlib
import random
import time
from importlib import import_module
from io import BytesIO
from pathlib import Path
from typing import Type

from PIL import Image

from spacer import config
from spacer.data_classes import PointFeatures, ImageFeatures
from spacer.exceptions import ConfigError, HashMismatchError
from spacer.extract_features_utils import crop_patches
from spacer.messages import DataLocation, ExtractFeaturesReturnMsg
from spacer.storage import storage_factory
from spacer.torch_utils import extract_feature


class FeatureExtractor(abc.ABC):

    # Subclasses should define their expected data_locations keys here.
    # See __init__() for an explanation of data_locations.
    DATA_LOCATION_KEYS: list[str] = []

    def __init__(self,
                 data_locations: dict[str, DataLocation],
                 data_hashes: dict[str, str] = None,
                 **kwargs):
        # DataLocations of files/blobs used to load the extractor.
        # Each subclass of FeatureExtractor should expect specific keys
        # to be present in this dict, one key per file/blob.
        self.data_locations = data_locations
        for key in self.DATA_LOCATION_KEYS:
            if key not in data_locations:
                raise ValueError(
                    f"{key} must be present in data_locations for"
                    f" {self.__class__.__name__}.")

        # SHA256 hashes checking the integrity of the extractor
        # files/blobs. The dict keys should match those of data_locations.
        # Don't have to specify a hash for every data location; only the
        # ones you want to check the integrity of.
        self.data_hashes = data_hashes or dict()

    @abc.abstractmethod
    def __call__(self,
                 im: Image,
                 rowcols: list[tuple[int, int]]) \
            -> tuple[ImageFeatures, ExtractFeaturesReturnMsg]:
        """ Runs the feature extraction """

    @property
    @abc.abstractmethod
    def feature_dim(self) -> int:
        """ Returns the feature dimension of extractor. """

    def serialize(self) -> dict:
        cls = self.__class__
        return dict(
            # Dotted path to the FeatureExtractor subclass,
            # e.g. 'spacer.extract_features.EfficientNetExtractor'
            class_path=f'{cls.__module__}.{cls.__name__}',
            data_locations=dict([
                (key, loc.serialize())
                for key, loc in self.data_locations.items()
            ]),
            data_hashes=self.data_hashes,
        )

    @staticmethod
    def deserialize(data: dict) -> 'FeatureExtractor':
        working_data = data.copy()
        class_path = working_data.pop('class_path')

        module_path, class_name = class_path.rsplit('.', 1)
        module = import_module(module_path)
        extractor_class: Type['FeatureExtractor'] = getattr(module, class_name)

        working_data['data_locations'] = dict([
            (key, DataLocation.deserialize(serialized_loc))
            for key, serialized_loc in working_data['data_locations'].items()
        ])

        return extractor_class(**working_data)

    def __repr__(self):
        """
        Seeing the serialized fields tends to be a more useful repr than
        `<SomeExtractor object>`.
        """
        return str(self.serialize())

    def __eq__(self, other):
        """This is for some tests comparing equality of messages."""
        return self.serialize() == other.serialize()

    def _check_data_hash(self, data, key):
        if key in self.data_hashes:
            sha256 = hashlib.sha256(data).hexdigest()
            if sha256 != self.data_hashes[key]:
                raise HashMismatchError(
                    f"Hash doesn't match for extractor's"
                    f" '{key}' file/blob.")

    def data_filepath_for_cache(self, key: str) -> Path:
        if not config.EXTRACTORS_CACHE_DIR:
            raise ConfigError(
                "Must define EXTRACTORS_CACHE_DIR"
                " to load extractor files/blobs into the filesystem.")

        loc = self.data_locations[key]
        return Path(config.EXTRACTORS_CACHE_DIR) / loc.filename

    def load_data_into_filesystem(self, key: str) -> tuple[str, bool]:
        """
        Ensure the data file/blob of the given key is loaded into the
        filesystem.
        Returns:
        - A filesystem path for that data file/blob.
        - True if it had to be loaded from remote storage, else False.
        """
        loc = self.data_locations[key]

        if loc.storage_type == 'filesystem':
            return loc.key, False

        filepath_for_cache = self.data_filepath_for_cache(key)
        file_storage = storage_factory('filesystem')
        already_in_filesystem = filepath_for_cache.exists()

        if not already_in_filesystem:
            # Must load into the filesystem.
            storage = storage_factory(loc.storage_type, loc.bucket_name)
            data = storage.load(loc.key)
            file_storage.store(str(filepath_for_cache), data)

        remote_loaded = not already_in_filesystem and loc.is_remote
        return str(filepath_for_cache), remote_loaded

    def load_data(self, key: str) -> tuple[BytesIO, bool]:
        """
        Loads extractor data from the DataLocation
        `self.data_locations[key]`.

        If the location is remote, caches the data to the filesystem.
        Cached files are identified by the source location's filename. So
        be sure that you don't have distinct files with the same filename,
        or they'll collide with each other.

        Returns:
        - Byte stream of the data file/blob.
        - True if it had to be loaded from remote storage, else False.

        Windows beware: behavior is undefined when there are distinct files
        to cache whose filenames differ only in upper/lowercase.
        """
        loc = self.data_locations[key]

        if loc.is_remote:

            data_filepath, remote_loaded = self.load_data_into_filesystem(key)
            file_storage = storage_factory('filesystem')
            data = file_storage.load(data_filepath)

            if remote_loaded:
                # Cache miss.
                check_hash = True
            else:
                # Cache hit.
                # Skip the hash check in this case, since presumably the
                # hash was already checked when loading into the cache, and
                # hash checking for large files may be time consuming.
                check_hash = False

        else:

            storage = storage_factory(loc.storage_type, loc.bucket_name)
            data = storage.load(loc.key)

            remote_loaded = False
            check_hash = True

        if check_hash:
            self._check_data_hash(data, key)
        return data, remote_loaded


class DummyExtractor(FeatureExtractor):
    """
    This doesn't actually extract any features from the image,
    it just returns dummy information.
    """
    def __init__(self,
                 data_locations: dict[str, DataLocation] = None,
                 feature_dim: int = 4096,
                 **kwargs):
        super().__init__(data_locations or dict(), **kwargs)

        # If you want these features to be compatible with an actual
        # classifier, you should make this match the feature_dim that
        # the classifier was trained on.
        self._feature_dim = feature_dim

    def __call__(self, im, rowcols):
        return ImageFeatures(
            point_features=[PointFeatures(row=rc[0],
                                          col=rc[1],
                                          data=[random.random() for _ in
                                                range(self.feature_dim)])
                            for rc in rowcols],
            valid_rowcol=True,
            npoints=len(rowcols),
            feature_dim=self.feature_dim
        ), ExtractFeaturesReturnMsg.example()

    @property
    def feature_dim(self):
        return self._feature_dim

    def serialize(self) -> dict:
        data = super().serialize()
        data['feature_dim'] = self.feature_dim
        return data


class VGG16CaffeExtractor(FeatureExtractor):

    # definition should be a Caffe prototxt file, typically .prototxt
    # weights should be a Caffe model file, typically .caffemodel
    DATA_LOCATION_KEYS = ['definition', 'weights']

    def __call__(self, im, rowcols):
        if not config.HAS_CAFFE:
            raise ConfigError(
                f"Need Caffe installed to call"
                f" {self.__class__.__name__}.")

        # We should only reach this line if it is confirmed caffe is available
        from spacer.caffe_utils import classify_from_patchlist

        start_time = time.time()

        # Set caffe parameters
        caffe_params = {'im_mean': [128, 128, 128],
                        'scaling_method': 'scale',
                        'crop_size': 224,
                        'batch_size': 10}

        # Crop patches
        with config.log_entry_and_exit('cropping of {} patches'.format(
                len(rowcols))):
            patch_list = crop_patches(im, rowcols, caffe_params['crop_size'])
        del im

        # Extract features
        definition_filepath, _ = \
            self.load_data_into_filesystem('definition')
        weights_filepath, remote_loaded = \
            self.load_data_into_filesystem('weights')

        feats = classify_from_patchlist(patch_list,
                                        caffe_params,
                                        definition_filepath,
                                        weights_filepath,
                                        scorelayer='fc7')

        return \
            ImageFeatures(
                point_features=[PointFeatures(row=rc[0],
                                              col=rc[1],
                                              data=ft.tolist())
                                for rc, ft in zip(rowcols, feats)],
                valid_rowcol=True,
                feature_dim=len(feats[0]),
                npoints=len(feats)
            ), ExtractFeaturesReturnMsg(
                extractor_loaded_remotely=remote_loaded,
                runtime=time.time() - start_time
            )

    @property
    def feature_dim(self):
        return 4096


class EfficientNetExtractor(FeatureExtractor):

    # weights should be a PyTorch tensor file, typically .pt
    DATA_LOCATION_KEYS = ['weights']

    def __call__(self, im, rowcols):

        start_time = time.time()

        weights_data, remote_loaded = self.load_data('weights')

        # Set torch parameters
        torch_params = {'model_type': 'efficientnet',
                        'model_name': 'efficientnet-b0',
                        'weights_data': weights_data,
                        'num_class': 1275,
                        'crop_size': 224,
                        'batch_size': 10}

        # Crop patches
        with config.log_entry_and_exit('cropping %s patches' % len(rowcols)):
            patch_list = crop_patches(im, rowcols, torch_params['crop_size'])
        del im

        # Extract features
        feats = extract_feature(patch_list, torch_params)

        return ImageFeatures(
            point_features=[PointFeatures(row=rc[0], col=rc[1], data=ft)
                            for rc, ft in zip(rowcols, feats)],
            valid_rowcol=True, feature_dim=len(feats[0]), npoints=len(feats)
        ), ExtractFeaturesReturnMsg(
            extractor_loaded_remotely=remote_loaded,
            runtime=time.time() - start_time
        )

    @property
    def feature_dim(self):
        return 1280



from typing import List, Tuple

import numpy as np
from PIL import Image


def gray2rgb(im: np.ndarray) -> np.ndarray:
    """
    Convert gray image to RGB image
    :param im: gray image to be converted
    :return: RGB image
    """
    w, h = im.shape
    ret = np.empty((w, h, 3), dtype=np.uint8)
    ret[:, :, 0] = im
    ret[:, :, 1] = im
    ret[:, :, 2] = im

    return ret


def crop_patches(im: Image,
                 rowcols: List[Tuple[int, int]],
                 crop_size: int) -> List[np.ndarray]:
    """
    Crop patches from an image
    :param im: image for cropping
    :param rowcols: [(row1, col1), (row2, col2), ...]
    :param crop_size: patch size
    :return: patch list
    """
    im = np.array(im)

    if len(im.shape) == 2 or im.shape[2] == 1:
        im = gray2rgb(im)
    im = im[:, :, :3]  # only keep the first three color channels

    pad = crop_size
    im = np.pad(im, ((pad, pad), (pad, pad), (0, 0)), mode='reflect')

    patches = [crop_simple(im, (row + pad, col + pad), crop_size)
               for row, col in rowcols]

    return patches


def crop_simple(im: np.ndarray,
                center: Tuple[int, int],
                crop_size: int) -> np.ndarray:
    """
    Crops an image around the given center
    :param im: image to be cropped
    :param center: offset (row, col)
    :param crop_size: cropping size
    :return: cropped image in numpy array
    """
    upper = int(center[0] - crop_size / 2)
    left = int(center[1] - crop_size / 2)

    return im[upper: upper + crop_size, left: left + crop_size, :]
